@@ -517,8 +517,8 @@ if rc != 0:
 if rc != 0:
     param = ""
 
-# On Mac we test for 'clang++' and use it by default
-if sys.platform == 'darwin':
+# On Mac/FreeBSD we test for 'clang++' and use it by default
+if sys.platform == 'darwin' or sys.platform.startswith('freebsd'):
     try:
         rc = call_subprocess_Popen(['clang++', '-v'])
         if rc == 0:
@@ -750,8 +750,9 @@ AddConfigVar('warn.ignore_bug_before',
               "bugs found after that version. "
               "Warning for specific bugs can be configured with specific "
               "[warn] flags."),
-             EnumStr('0.8', 'None', 'all', '0.3', '0.4', '0.4.1', '0.5', '0.6',
-                     '0.7', '0.8', '0.8.1', '0.8.2', '0.9', '0.10', '1.0', '1.0.1',
+             EnumStr('0.9', 'None', 'all', '0.3', '0.4', '0.4.1', '0.5', '0.6',
+                     '0.7', '0.8', '0.8.1', '0.8.2', '0.9', '0.10', '1.0',
+                     '1.0.1', '1.0.2',
                      allow_override=False),
              in_c_key=False)
 
@@ -1246,14 +1247,24 @@ def check_mkl_openmp():
     if ('MKL_THREADING_LAYER' in os.environ and
             os.environ['MKL_THREADING_LAYER'] == 'GNU'):
         return
+    try:
+        import numpy._mklinit  # noqa
+        return
+    except ImportError:
+        pass
     try:
         import mkl
         if '2018' in mkl.get_version_string():
-            raise RuntimeError('To use MKL 2018 with Theano you MUST set "MKL_THREADING_LAYER=GNU" in your environement.')
+            raise RuntimeError("""
+To use MKL 2018 with Theano either update the numpy conda packages to
+their latest build or set "MKL_THREADING_LAYER=GNU" in your
+environment.
+""")
     except ImportError:
         raise RuntimeError("""
-Could not import 'mkl'.  Either install mkl-service with conda or set
-MKL_THREADING_LAYER=GNU in your environment for MKL 2018.
+Could not import 'mkl'.  If you are using conda, update the numpy
+packages to the latest build otherwise, set MKL_THREADING_LAYER=GNU in
+your environment for MKL 2018.
 
 If you have MKL 2017 install and are not in a conda environment you
 can set the Theano flag blas.check_openmp to False.  Be warned that if
@@ -1394,13 +1405,23 @@ def default_blas_ldflags():
                 thr = 'mkl_gnu_thread'
             else:
                 thr = 'mkl_intel_thread'
+            base_flags = list(flags)
             flags += ['-l%s' % l for l in ["mkl_core",
                                            thr,
                                            "mkl_rt"]]
             res = try_blas_flag(flags)
+
+            if not res and sys.platform == "win32" and thr == "mkl_gnu_thread":
+                # Check if it would work for intel OpenMP on windows
+                flags = base_flags + ['-l%s' % l for l in ["mkl_core",
+                                                           'mkl_intel_thread',
+                                                           "mkl_rt"]]
+                res = try_blas_flag(flags)
+
             if res:
                 check_mkl_openmp()
                 return res
+
             flags.extend(['-Wl,-rpath,' + l for l in
                           blas_info.get('library_dirs', [])])
             res = try_blas_flag(flags)
