@@ -1,5 +1,5 @@
 """
-This module provides the Scan Op
+This module provides the Scan Op.
 
 Scanning is a general form of recurrence, which can be used for looping.
 The idea is that you *scan* a function along some input sequence, producing
@@ -32,7 +32,9 @@ host at each step
 The Scan Op should typically be used by calling any of the following
 functions: ``scan()``, ``map()``, ``reduce()``, ``foldl()``,
 ``foldr()``.
+
 """
+from __future__ import absolute_import, print_function, division
 __docformat__ = 'restructedtext en'
 __authors__ = ("Razvan Pascanu "
                "Frederic Bastien "
@@ -42,11 +44,14 @@ __copyright__ = "(c) 2010, Universite de Montreal"
 __contact__ = "Razvan Pascanu <r.pascanu@gmail>"
 
 
-import itertools
 import logging
-import numpy
+import numpy as np
 import warnings
+from collections import OrderedDict
 
+from theano.compat import ifilter, izip
+from six import iteritems, integer_types
+from six.moves import xrange
 from theano.compile import SharedVariable, function
 from theano import compile
 from theano import gof
@@ -55,7 +60,6 @@ from theano import tensor
 from theano import config
 from theano.updates import OrderedUpdates
 from theano.compile import ops
-from theano.compat.python2x import OrderedDict
 
 
 from theano.scan_module import scan_op
@@ -77,12 +81,14 @@ def scan(fn,
          name=None,
          profile=False,
          allow_gc=None,
-         strict=False):
-    """
-    This function constructs and applies a Scan op to the provided
+         strict=False,
+         return_list=False):
+    """This function constructs and applies a Scan op to the provided
     arguments.
 
-    :param fn:
+    Parameters
+    ----------
+    fn
         ``fn`` is a function that describes the operations involved in one
         step of ``scan``. ``fn`` should construct variables describing the
         output of one iteration step. It should expect as input theano
@@ -173,7 +179,7 @@ def scan(fn,
         number of steps ) is still required even though a condition is
         passed (and it is used to allocate memory if needed). = {}):
 
-    :param sequences:
+    sequences
         ``sequences`` is the list of Theano variables or dictionaries
         describing the sequences ``scan`` has to iterate over. If a
         sequence is given as wrapped in a dictionary, then a set of optional
@@ -191,8 +197,7 @@ def scan(fn,
         Any Theano variable in the list ``sequences`` is automatically
         wrapped into a dictionary where ``taps`` is set to ``[0]``
 
-
-    :param outputs_info:
+    outputs_info
         ``outputs_info`` is the list of Theano variables or dictionaries
         describing the initial state of the outputs computed
         recurrently. When this initial states are given as dictionary
@@ -250,26 +255,21 @@ def scan(fn,
         raised (because there is no convention on how scan should map
         the provided information to the outputs of ``fn``)
 
-
-    :param non_sequences:
+    non_sequences
         ``non_sequences`` is the list of arguments that are passed to
         ``fn`` at each steps. One can opt to exclude variable
         used in ``fn`` from this list as long as they are part of the
         computational graph, though for clarity we encourage not to do so.
 
-
-    :param n_steps:
+    n_steps
         ``n_steps`` is the number of steps to iterate given as an int
         or Theano scalar. If any of the input sequences do not have
         enough elements, scan will raise an error. If the *value is 0* the
-        outputs will have *0 rows*. If the value is negative, ``scan``
-        will run backwards in time. If the ``go_backwards`` flag is already
-        set and also ``n_steps`` is negative, ``scan`` will run forward
-        in time. If n_steps is not provided, ``scan`` will figure
-        out the amount of steps it should run given its input sequences.
+        outputs will have *0 rows*. If n_steps is not provided, ``scan`` will
+        figure out the amount of steps it should run given its input
+        sequences. ``n_steps`` < 0 is not supported anymore.
 
-
-    :param truncate_gradient:
+    truncate_gradient
         ``truncate_gradient`` is the number of steps to use in truncated
         BPTT.  If you compute gradients through a scan op, they are
         computed using backpropagation through time. By providing a
@@ -277,16 +277,14 @@ def scan(fn,
         of classical BPTT, where you go for only ``truncate_gradient``
         number of steps back in time.
 
-
-    :param go_backwards:
+    go_backwards
         ``go_backwards`` is a flag indicating if ``scan`` should go
         backwards through the sequences. If you think of each sequence
         as indexed by time, making this flag True would mean that
         ``scan`` goes back in time, namely that for any sequence it
         starts from the end and goes towards 0.
 
-
-    :param name:
+    name
         When profiling ``scan``, it is crucial to provide a name for any
         instance of ``scan``. The profiler will produce an overall
         profile of your code as well as profiles for the computation of
@@ -294,7 +292,7 @@ def scan(fn,
         appears in those profiles and can greatly help to disambiguate
         information.
 
-    :param mode:
+    mode
         It is recommended to leave this argument to None, especially
         when profiling ``scan`` (otherwise the results are not going to
         be accurate). If you prefer the computations of one step of
@@ -303,7 +301,7 @@ def scan(fn,
         loop are done (see ``theano.function`` for details about
         possible values and their meaning).
 
-    :param profile:
+    profile
         Flag or string. If true, or different from the empty string, a
         profile object will be created and attached to the inner graph of
         scan. In case ``profile`` is True, the profile object will have the
@@ -312,39 +310,60 @@ def scan(fn,
         inner graph with the new cvm linker ( with default modes,
         other linkers this argument is useless)
 
-    :param allow_gc:
+    allow_gc
         Set the value of allow gc for the internal graph of scan.  If
         set to None, this will use the value of config.scan.allow_gc.
 
-    :param strict:
+        The full scan behavior related to allocation is determined by
+        this value and the Theano flag allow_gc. If the flag allow_gc
+        is True (default) and this scan parameter allow_gc is False
+        (default), then we let scan allocate all intermediate memory
+        on the first iteration, those are not garbage collected them
+        during that first iteration (this is determined by the scan
+        allow_gc). This speed up allocation of the following
+        iteration. But we free all those temp allocation at the end of
+        all iterations (this is what the Theano flag allow_gc mean).
+
+        If you use preallocate and this scan is on GPU, the speed up
+        from the scan allow_gc is small. If you are missing memory,
+        disable the scan allow_gc could help you run graph that
+        request much memory.
+
+    strict
         If true, all the shared variables used in ``fn`` must be provided as a
-        part of ``non_sequences`` or ``sequences``. 
+        part of ``non_sequences`` or ``sequences``.
+
+    return_list
+        If True, will always return a list, even if there is only 1 output.
+
+    Returns
+    -------
+    tuple
+        Tuple of the form (outputs, updates); ``outputs`` is either a
+        Theano variable or a list of Theano variables representing the
+        outputs of ``scan`` (in the same order as in ``outputs_info``).
+        ``updates`` is a subclass of dictionary specifying the update rules for
+        all shared variables used in scan.
+        This dictionary should be passed to ``theano.function`` when you compile
+        your function. The change compared to a normal dictionary is that we
+        validate that keys are SharedVariable and addition of those dictionary
+        are validated to be consistent.
 
-    :rtype: tuple
-    :return: tuple of the form (outputs, updates); ``outputs`` is either a
-             Theano variable or a list of Theano variables representing the
-             outputs of ``scan`` (in the same order as in
-             ``outputs_info``). ``updates`` is a subclass of dictionary
-             specifying the
-             update rules for all shared variables used in scan
-             This dictionary should be passed to ``theano.function`` when
-             you compile your function. The change compared to a normal
-             dictionary is that we validate that keys are SharedVariable
-             and addition of those dictionary are validated to be consistent.
     """
     # General observation : this code is executed only once, at creation
     # of the computational graph, so we don't yet need to be smart about
     # anything (to speed things up)
 
     ##
-    ###   Step 1. Wrap all inputs in dictionaries and add default values
+    # Step 1. Wrap all inputs in dictionaries and add default values
     ##
 
     # check if inputs are just single variables instead of lists
     def wrap_into_list(x):
-        '''
-        Wrap the input into a list if it is not already a list
-        '''
+        """
+        Wrap the input into a list if it is not already a list.
+
+        """
         if x is None:
             return []
         elif not isinstance(x, (list, tuple)):
@@ -370,7 +389,7 @@ def scan(fn,
     # To do that we check here to see the nature of n_steps
     n_fixed_steps = None
 
-    if isinstance(n_steps, (float, int)):
+    if isinstance(n_steps, (float, integer_types)):
         n_fixed_steps = int(n_steps)
     else:
         try:
@@ -380,7 +399,7 @@ def scan(fn,
 
     # Check n_steps is an int
     if (hasattr(n_steps, 'dtype') and
-        str(n_steps.dtype)[:3] not in ('uin', 'int')):
+        str(n_steps.dtype) not in tensor.integer_dtypes):
         raise ValueError(' n_steps must be an int. dtype provided '
                          'is %s' % n_steps.dtype)
 
@@ -431,14 +450,24 @@ def scan(fn,
                              getattr(outs_info[i]['initial'], 'name', 'None'),
                              i)
                 outs_info[i]['taps'] = [-1]
+            elif outs_info[i].get('taps', None) is not None:
+                # Check that taps are valid (< 0 and all dfferent)
+                taps = outs_info[i]['taps']
+                if len(taps) > len(set(taps)):
+                    raise ValueError(('All the taps must be different in '
+                                      ' `outputs_info`'), outs_info[i])
+                for t in taps:
+                    if t >= 0:
+                        raise ValueError(('All the tap values must be '
+                                          'smaller than 0.'), outs_info[i])
         else:
             # if a None is provided as the output info we replace it
             # with an empty OrdereDict() to simplify handling
             outs_info[i] = OrderedDict()
 
     ##
-    ###   Step 2. Generate inputs and outputs of the inner functions
-    ###           for compiling a dummy function (Iteration #1)
+    # Step 2. Generate inputs and outputs of the inner functions
+    # for compiling a dummy function (Iteration #1)
     ##
 
     # create theano inputs for the recursive function
@@ -458,8 +487,14 @@ def scan(fn,
         # a sequence, though is highly unlikely in practice
         if 'taps' in seq:
             # go through the indicated slice
-            mintap = numpy.min(seq['taps'])
-            maxtap = numpy.max(seq['taps'])
+            mintap = np.min(seq['taps'])
+            maxtap = np.max(seq['taps'])
+            # We cut the sequence such that seq[i] to correspond to
+            # seq[i-k]. For the purposes of cutting the sequences, we
+            # need to pretend tap 0 is used to avoid cutting the sequences
+            # too long if the taps are all lower or all higher than 0.
+            maxtap_proxy = max(maxtap, 0)
+            mintap_proxy = min(mintap, 0)
             for k in seq['taps']:
                 # create one slice of the input
                 # Later on, if we decide not to use scan because we are
@@ -470,9 +505,9 @@ def scan(fn,
 
                 # If not we need to use copies, that will be replaced at
                 # each frame by the corresponding slice
-                actual_slice = seq['input'][k - mintap]
+                actual_slice = seq['input'][k - mintap_proxy]
                 _seq_val = tensor.as_tensor_variable(seq['input'])
-                _seq_val_slice = _seq_val[k - mintap]
+                _seq_val_slice = _seq_val[k - mintap_proxy]
                 nw_slice = _seq_val_slice.type()
 
                 # Try to transfer test_value to the new variable
@@ -480,7 +515,7 @@ def scan(fn,
                     try:
                         nw_slice.tag.test_value = gof.Op._get_test_value(
                             _seq_val_slice)
-                    except AttributeError, e:
+                    except AttributeError as e:
                         if config.compute_test_value != 'ignore':
                             # No need to print a warning or raise an error now,
                             # it will be done when fn will be called.
@@ -499,21 +534,18 @@ def scan(fn,
                         nw_name = seq['input'].name + '[t%d]' % k
                     nw_slice.name = nw_name
 
-                # We cut the sequence such that seq[i] to correspond to
-                # seq[i-k]
-                if maxtap < 0:
-                    offset = abs(maxtap)
+                start = (k - mintap_proxy)
+                nw_name = None
+                if k == maxtap_proxy:
+                    nw_seq = seq['input'][start:]
+                    if getattr(seq['input'], 'name', None) is not None:
+                        nw_name = seq['input'].name + "[%d:]" % start
                 else:
-                    offset = 0
-                if maxtap == mintap and maxtap != 0:
-                    if maxtap < 0:
-                        nw_seq = seq['input'][:maxtap]
-                    else:
-                        nw_seq = seq['input'][maxtap:]
-                elif maxtap - k != 0:
-                    nw_seq = seq['input'][offset + k - mintap: -(maxtap - k)]
-                else:
-                    nw_seq = seq['input'][offset + k - mintap:]
+                    end = -(maxtap_proxy - k)
+                    nw_seq = seq['input'][start:end]
+                    if getattr(seq['input'], 'name', None) is not None:
+                        nw_name = seq['input'].name + "[%d:%d]" % (start, end)
+
                 if go_backwards:
                     nw_seq = nw_seq[::-1]
 
@@ -521,6 +553,9 @@ def scan(fn,
                 inner_seqs.append(nw_slice)
                 inner_slices.append(actual_slice)
                 n_seqs += 1
+                # Add names -- it helps a lot when debugging
+                if nw_name is not None:
+                    nw_seq.name = nw_name
 
     # Since we've added all sequences now we need to level them up based on
     # n_steps or their different shapes
@@ -534,7 +569,7 @@ def scan(fn,
 
     if len(lengths_vec) == 0:
         # ^ No information about the number of steps
-        raise ValueError(' No information about the number of steps '
+        raise ValueError('No information about the number of steps '
                          'provided. Either provide a value for '
                          'n_steps argument of scan or provide an input '
                          'sequence')
@@ -548,12 +583,6 @@ def scan(fn,
     else:
         actual_n_steps = tensor.as_tensor(n_steps)
 
-    # Add names -- it helps a lot when debugging
-
-    for (nw_seq, seq) in zip(scan_seqs, seqs):
-        if getattr(seq['input'], 'name', None) is not None:
-            nw_seq.name = seq['input'].name + '[%d:]' % k
-
     scan_seqs = [seq[:actual_n_steps] for seq in scan_seqs]
     # Conventions :
     #   mit_mot = multiple input taps, multiple output taps ( only provided
@@ -612,7 +641,7 @@ def scan(fn,
             if config.compute_test_value != 'off':
                 try:
                     arg.tag.test_value = gof.Op._get_test_value(actual_arg)
-                except AttributeError, e:
+                except AttributeError as e:
                     if config.compute_test_value != 'ignore':
                         # No need to print a warning or raise an error now,
                         # it will be done when fn will be called.
@@ -627,7 +656,7 @@ def scan(fn,
             # the initial state over. We do this using the expand function
             # defined in scan utils
             sit_sot_scan_inputs.append(
-                scan_utils.expand(
+                scan_utils.expand_empty(
                     tensor.unbroadcast(
                         tensor.shape_padleft(actual_arg), 0),
                     actual_n_steps
@@ -642,19 +671,19 @@ def scan(fn,
 
         elif init_out.get('taps', None):
 
-            if numpy.any(numpy.array(init_out.get('taps', [])) > 0):
+            if np.any(np.array(init_out.get('taps', [])) > 0):
                 # Make sure we do not have requests for future values of a
                 # sequence we can not provide such values
                 raise ValueError('Can not use future taps of outputs',
                                     init_out)
             # go through the taps
-            mintap = abs(numpy.min(init_out['taps']))
+            mintap = abs(np.min(init_out['taps']))
             mit_sot_tap_array.append(init_out['taps'])
-            idx_offset = abs(numpy.min(init_out['taps']))
+            idx_offset = abs(np.min(init_out['taps']))
             # Sequence
             mit_sot_scan_inputs.append(
-                scan_utils.expand(init_out['initial'][:mintap],
-                                 actual_n_steps))
+                scan_utils.expand_empty(init_out['initial'][:mintap],
+                                        actual_n_steps))
 
             if i in return_steps:
                 mit_sot_return_steps[n_mit_sot] = return_steps[i]
@@ -672,7 +701,7 @@ def scan(fn,
                     try:
                         nw_slice.tag.test_value = gof.Op._get_test_value(
                             _init_out_var_slice)
-                    except AttributeError, e:
+                    except AttributeError as e:
                         if config.compute_test_value != 'ignore':
                             # No need to print a warning or raise an error now,
                             # it will be done when fn will be called.
@@ -692,14 +721,14 @@ def scan(fn,
                                             '[t%d]' % k)
                 mit_sot_inner_inputs.append(nw_slice)
                 mit_sot_inner_slices.append(actual_nw_slice)
-        #NOTE: there is another case, in which we do not want to provide
+        # NOTE: there is another case, in which we do not want to provide
         #      any previous value of the output to the inner function (i.e.
         #      a map); in that case we do not have to do anything ..
 
     # Re-order args
-    max_mit_sot = numpy.max([-1] + mit_sot_rightOrder) + 1
-    max_sit_sot = numpy.max([-1] + sit_sot_rightOrder) + 1
-    n_elems = numpy.max([max_mit_sot, max_sit_sot])
+    max_mit_sot = np.max([-1] + mit_sot_rightOrder) + 1
+    max_sit_sot = np.max([-1] + sit_sot_rightOrder) + 1
+    n_elems = np.max([max_mit_sot, max_sit_sot])
     _ordered_args = [[] for x in xrange(n_elems)]
     offset = 0
     for idx in xrange(n_mit_sot):
@@ -748,7 +777,7 @@ def scan(fn,
     else:
         as_while = False
     ##
-    ###   Step 3. Check if we actually need scan and remove it if we don't
+    # Step 3. Check if we actually need scan and remove it if we don't
     ##
 
     if n_fixed_steps in [1, -1]:
@@ -770,13 +799,14 @@ def scan(fn,
                 return_steps.get(pos, 0) != 1):
                 outputs[pos] = tensor.unbroadcast(
                     tensor.shape_padleft(inner_out), 0)
-        if len(outputs) == 1:
+
+        if return_list is not True and len(outputs) == 1:
             outputs = outputs[0]
 
         return (outputs, updates)
 
     ##
-    ###   Step 4. Compile the dummy function
+    # Step 4. Compile the dummy function
     ##
 
     # We can now compile a dummy function just to see what shared variable
@@ -789,38 +819,45 @@ def scan(fn,
 
     # extract still missing inputs (there still might be so) and add them
     # as non sequences at the end of our args
+    if condition is not None:
+        outputs.append(condition)
     fake_nonseqs = [x.type() for x in non_seqs]
     fake_outputs = scan_utils.clone(outputs,
-                                    replace=OrderedDict(zip(non_seqs,
-                                                     fake_nonseqs)))
-    all_inputs = itertools.ifilter(
+                                    replace=OrderedDict(izip(non_seqs,
+                                                             fake_nonseqs)))
+    all_inputs = ifilter(
         lambda x: (isinstance(x, gof.Variable) and
                    not isinstance(x, SharedVariable) and
                    not isinstance(x, gof.Constant)),
         gof.graph.inputs(fake_outputs))
     extra_inputs = [x for x in all_inputs if x not in args + fake_nonseqs]
     non_seqs += extra_inputs
-    ## Note we do not use all_inputs directly since the order of variables
-    ## in args is quite important
+    # Note we do not use all_inputs directly since the order of variables
+    # in args is quite important
     dummy_args += extra_inputs
 
     dummy_outs = outputs
-    if condition is not None:
-        dummy_outs.append(condition)
-    dummy_f = function(dummy_args,
-                       dummy_outs,
-                       updates=updates,
-                       mode=compile.mode.Mode(linker='py',
-                                              optimizer=None),
-                       on_unused_input='ignore',
-                       profile=False)
-
+    # Perform a try-except to provide a meaningful error message to the
+    # user if inputs of the inner function are missing.
+    try:
+        dummy_f = function(dummy_args,
+                           dummy_outs,
+                           updates=updates,
+                           mode=compile.mode.Mode(linker='py',
+                                                  optimizer=None),
+                           on_unused_input='ignore',
+                           profile=False)
+    except gof.fg.MissingInputError as err:
+        msg = ("\nPlease pass this variable to the scan's inner function. Do "
+               "not forget to also pass it to the `non_sequences` attribute "
+               "of scan.")
+        raise gof.fg.MissingInputError(err.args[0] + msg)
     ##
-    ### Step 5. Re-arange inputs of scan into a more strict order
+    # Step 5. Re-arange inputs of scan into a more strict order
     ##
 
-    ## Step 5.0 Check the outputs of the dummy function to see if they
-    ##          match with user provided data
+    # Step 5.0 Check the outputs of the dummy function to see if they
+    # match with user provided data
 
     # if the number of outputs to the function does not match the number of
     # assumed outputs until now (provided by the user) there can be
@@ -840,18 +877,18 @@ def scan(fn,
             n_outs = n_outs - 1
         outs_info = [OrderedDict() for x in xrange(n_outs)]
 
-    ## Step 5.1 Outputs with taps different then -1
+    # Step 5.1 Outputs with taps different then -1
 
     for i, out in enumerate(outs_info):
         if 'taps' in out and out['taps'] != [-1]:
             mit_sot_inner_outputs.append(outputs[i])
 
-    ## Step 5.2 Outputs with tap equal to -1
+    # Step 5.2 Outputs with tap equal to -1
     for i, out in enumerate(outs_info):
         if 'taps' in out and out['taps'] == [-1]:
             sit_sot_inner_outputs.append(outputs[i])
 
-    ## Step 5.3 Outputs that correspond to update rules of shared variables
+    # Step 5.3 Outputs that correspond to update rules of shared variables
     givens = OrderedDict()
     n_shared_outs = 0
     shared_scan_inputs = []
@@ -866,11 +903,12 @@ def scan(fn,
             if isinstance(new_var.type, ops.expandable_types):
                 sit_sot_inner_inputs.append(new_var)
                 sit_sot_scan_inputs.append(
-                    scan_utils.expand(
+                    scan_utils.expand_empty(
                         tensor.unbroadcast(
                             tensor.shape_padleft(input.variable), 0),
                         actual_n_steps))
-                sit_sot_inner_outputs.append(input.update)
+                tensor_update = tensor.as_tensor_variable(input.update)
+                sit_sot_inner_outputs.append(tensor_update)
                 # Not that pos is not a negative index. The sign of pos is used
                 # as a flag to indicate if this output should be part of the
                 # update rules or part of the standard outputs of scan.
@@ -889,7 +927,7 @@ def scan(fn,
                 givens[input.variable] = new_var
                 n_shared_outs += 1
     n_sit_sot = len(sit_sot_inner_inputs)
-    ## Step 5.4 Outputs with no taps used in the input
+    # Step 5.4 Outputs with no taps used in the input
     n_nit_sot = 0
     nit_sot_inner_outputs = []
     nit_sot_return_steps = OrderedDict()
@@ -902,7 +940,7 @@ def scan(fn,
             nit_sot_rightOrder.append(i)
             n_nit_sot += 1
 
-    ## Step 5.5 all other arguments including extra inputs
+    # Step 5.5 all other arguments including extra inputs
     other_scan_args = []
     other_inner_args = []
 
@@ -910,15 +948,15 @@ def scan(fn,
                         if (not isinstance(arg, SharedVariable) and
                             not isinstance(arg, tensor.Constant))]
 
-    ## Step 5.6 all shared variables with no update rules
+    # Step 5.6 all shared variables with no update rules
     other_inner_args += [safe_new(arg, '_copy') for arg in non_seqs
                          if (not isinstance(arg, SharedVariable) and
                              not isinstance(arg, tensor.Constant))]
 
-    givens.update(OrderedDict(zip(other_scan_args, other_inner_args)))
+    givens.update(OrderedDict(izip(other_scan_args, other_inner_args)))
 
     if strict:
-        non_seqs_set = set(non_sequences if non_sequences != None else [])
+        non_seqs_set = set(non_sequences if non_sequences is not None else [])
 
         other_shared_scan_args = [arg.variable for arg
                             in dummy_f.maker.expanded_inputs
@@ -939,12 +977,12 @@ def scan(fn,
                             in dummy_f.maker.expanded_inputs
                             if (isinstance(arg.variable, SharedVariable) and
                                 not arg.update)]
-    givens.update(OrderedDict(zip(other_shared_scan_args,
-                           other_shared_inner_args)))
+    givens.update(OrderedDict(izip(other_shared_scan_args,
+                                   other_shared_inner_args)))
 
     ##
-    ### Step 6. Re-order the outputs and clone them replacing things
-    ###         using the givens
+    # Step 6. Re-order the outputs and clone them replacing things
+    # using the givens
     ##
     inner_inputs = (inner_seqs +
                     mit_mot_inner_inputs +
@@ -961,21 +999,21 @@ def scan(fn,
                   shared_inner_outputs)
     if condition is not None:
         inner_outs.append(condition)
-    # Cuda is imported here, instead of being imported on top of the file
-    # because forces on the user some dependencies that we might do not want
-    # to. Currently we are working on removing the dependencies on sandbox
-    # code completeley.
-    from theano.sandbox import cuda
-    if cuda.cuda_available:
+    # gpuarray is imported here, instead of being imported on top of
+    # the file because that would force on the user some dependencies that we
+    # might do not want to. Currently we are working on removing the
+    # dependencies on sandbox code completeley.
+    from theano import gpuarray
+    if gpuarray.pygpu_activated:
         # very often we end up in this situation when we want to
-        # replace w with w_copy, where w is CudaNdarray
+        # replace w with w_copy, where w is a GPU variable
         # and w_copy is TensorType. This is caused because shared
-        # variables are put on GPU right aways >:| ,
+        # variables are put on GPU right away >:| ,
         new_givens = OrderedDict()
 
-        for w, w_copy in givens.iteritems():
-            if (isinstance(w.type, cuda.CudaNdarrayType)
-                and isinstance(w_copy.type, tensor.TensorType)):
+        for w, w_copy in iteritems(givens):
+            if (isinstance(w.type, gpuarray.GpuArrayType) and
+                isinstance(w_copy.type, tensor.TensorType)):
                 for o in inner_outs:
                     new_givens = traverse(o, w, w_copy, new_givens)
             else:
@@ -986,7 +1024,7 @@ def scan(fn,
     new_outs = scan_utils.clone(inner_outs, replace=new_givens)
 
     ##
-    ### Step 7. Create the Scan Op
+    # Step 7. Create the Scan Op
     ##
 
     tap_array = mit_sot_tap_array + [[-1] for x in xrange(n_sit_sot)]
@@ -1007,19 +1045,16 @@ def scan(fn,
     info['name'] = name
     info['mode'] = mode
     info['destroy_map'] = OrderedDict()
-    info['gpu'] = False
+    info['gpua'] = False
     info['as_while'] = as_while
     info['profile'] = profile
     info['allow_gc'] = allow_gc
     info['strict'] = strict
-    if strict:
-        warnings.warn('In the strict mode, all neccessary shared variables '
-                      'must be passed as a part of non_sequences', Warning)
 
     local_op = scan_op.Scan(inner_inputs, new_outs, info)
 
     ##
-    ### Step 8. Compute the outputs using the scan op
+    # Step 8. Compute the outputs using the scan op
     ##
     _scan_inputs = (scan_seqs +
                     mit_mot_scan_inputs +
@@ -1036,15 +1071,15 @@ def scan(fn,
             arg = tensor.as_tensor_variable(arg)
         except TypeError:
             # This happens for Random States for e.g. but it is a good way
-            # to make sure no input is a cuda ndarrays
+            # to make sure all inputs are tensors.
             pass
         scan_inputs += [arg]
     scan_outs = local_op(*scan_inputs)
     if type(scan_outs) not in (list, tuple):
         scan_outs = [scan_outs]
     ##
-    ### Step 9. Figure out which outs are update rules for shared variables
-    ###         and so on ...
+    # Step 9. Figure out which outs are update rules for shared variables
+    # and so on ...
     ##
 
     update_map = OrderedUpdates()
@@ -1065,7 +1100,7 @@ def scan(fn,
         return out_ls
 
     offset = n_mit_mot
-    offsets = [abs(numpy.min(x)) for x in mit_sot_tap_array]
+    offsets = [abs(np.min(x)) for x in mit_sot_tap_array]
     mit_sot_outs = remove_dimensions(
         scan_outs[offset:offset + n_mit_sot],
         mit_sot_return_steps,
@@ -1110,8 +1145,9 @@ def scan(fn,
             # refers to update rule of index -1 - `pos`.
             update_map[sit_sot_shared[abs(pos) - 1]] = _scan_out_list[idx][-1]
     scan_out_list = [x for x in scan_out_list if x is not None]
-    if len(scan_out_list) == 1:
+    if return_list is not True and len(scan_out_list) == 1:
         scan_out_list = scan_out_list[0]
     elif len(scan_out_list) == 0:
         scan_out_list = None
+
     return (scan_out_list, update_map)
